from ast import List
import pickle
from endoreg_db.models import Requirement, Examination, PatientExamination
from endoreg_db.serializers import PatientExaminationSerializer, ExaminationSerializer, FindingClassificationSerializer, RequirementSerializer


class LookupService:
    def __init__(self, db_session):
        self.db_session = db_session
        self.hexdict = int()

    def hash_lookup(self, key, value):
        """
        Store a key-value pair in the lookup dictionary.
        """
        lookup_dict = pickle.loads(bytes.fromhex(self.hexdict))
        lookup_dict[key] = value
        self.hexdict = pickle.dumps(dict).hex()
        return True
    def hash_retrieve(self, key):
        """
        Retrieve a value by key from the lookup dictionary.
        """
        lookup_dict = pickle.loads(bytes.fromhex(self.hexdict))
        return lookup_dict.get(key, None)
    
    def hash_add_key_value(self, key, value):
        """
        Add a key-value pair to the lookup dictionary.
        """
        lookup_dict = pickle.loads(bytes.fromhex(self.hexdict))
        lookup_dict[key] = value
        self.hexdict = pickle.dumps(lookup_dict).hex()
        return True
    
    def hash_dehash(self):
        """
        Retrieve the entire lookup dictionary.
        """
        lookup_dict = pickle.loads(bytes.fromhex(self.hexdict))
        return lookup_dict
    
    def fetch_examination_dict(self, patient_id, tags: List[str] = None):
        """
        Fetch examinations for a given patient ID, optionally filtered by tags.
        """
        query = self.db_session.query(Examination).filter(Examination.patient_id == patient_id)
        if tags:
            query = query.filter(Examination.tags.any(Examination.tags.in_(tags)))
        examinations = query.all()
        return [ExaminationSerializer(exam).data for exam in examinations]  
    
    def fetch_patient_examination_by_id(self, patient_id, examination_id):
        """
        Fetch a patient examination object from the database by its ID.
        """
        return self.db_session.query(PatientExamination).filter(PatientExamination.id == patient_examination_id).first()
    
    def fetch_patient_examinations(self, patient_id):
        """
        Fetch all patient examinations for a given patient ID.
        """
        examinations = self.db_session.query(PatientExamination).filter(PatientExamination.patient_id == patient_id).all()
        return [PatientExaminationSerializer(exam).data for exam in examinations]

    def fetch_findings_by_patient_examination(self, patient_examination_id):
        """
        Fetch findings associated with a given patient examination ID.
        """
        patient_examination = self.db_session.query(PatientExamination).filter(PatientExamination.id == patient_examination_id).first()
        if not patient_examination:
            return []
        return [finding.serialize() for finding in patient_examination.findings]
    
    def fetch_classifications_by_finding(self, finding_id):
        """
        Fetch classifications associated with a given finding ID.
        """
        finding = self.db_session.query(Finding).filter(Finding.id == finding_id).first()
        if not finding:
            return []
        return [classification.serialize() for classification in finding.classifications]
    
    def fetch_default_finding_classification(self):
        """
        Fetch the default finding classification from the database.
        """
        return self.db_session.query(FindingClassification).filter(FindingClassification.is_default == True).first()
    
    def fetch_finding_classification_choice_by_requirement(self, requirement_id):
        """
        Retrieve finding classification choices based on the provided requirement ID.
        """
        requirement = self.get_requirement_by_id(requirement_id)
        if not requirement:
            return []
        return [choice.serialize() for choice in requirement.finding_classification_choices]
    
    def fetch_requirement_choices_by_requirement(self, requirement_id):
        """
        Retrieve requirement choices based on the provided requirement ID.
        """
        requirement = self.get_requirement_by_id(requirement_id)
        if not requirement:
            return []
        return [choice.serialize() for choice in requirement.requirement_choices]
    
    def available_finding_classification_choices_by_finding_classification(self, finding_classification):
        """
        Retrieve available finding classification choices based on the provided finding classification.
        """
        return self.db_session.query(FindingClassificationChoice).filter(
            FindingClassificationChoice.finding_classification_id == finding_classification.id
        ).all()

    def get_requirement_by_id(self, requirement_id):
        """
        Fetch a requirement object from the database by its ID.
        """
        return self.db_session.query(Requirement).filter(Requirement.id == requirement_id).first()

    def get_all_requirements_for_requirement_sets(self, requirement_set_ids: List[int]):
        """
        Fetch all requirements associated with the given requirement set IDs.
        """
        return self.db_session.query(Requirement).filter(Requirement.requirement_set_id.in_(requirement_set_ids)).all()
    
    def get_all_requirement_sets_for_examination(self, examination_id):
        """
        Fetch all requirement sets associated with the given examination ID.
        """
        examination = self.db_session.query(Examination).filter(Examination.id == examination_id).first()
        if not examination:
            return []
        return [req_set.serialize() for req_set in examination.requirement_sets]

    def serialize_requirement(self, requirement):
        """
        Serialize a requirement object to a byte stream.
        """
        return pickle.dumps(requirement)

    def deserialize_requirement(self, serialized_requirement):
        """
        Deserialize a byte stream back to a requirement object.
        """
        return pickle.loads(serialized_requirement)


from typing import TYPE_CHECKING, List, Optional, cast  # Added List

from django.db import models

# Added imports for type hints
if TYPE_CHECKING:
    from ....utils.links.requirement_link import RequirementLinks  # Added RequirementLinks
    from ...administration.person.patient import Patient
    from ...other.unit import Unit
    from ..medication import Medication, MedicationIndication, MedicationIntakeTime


class PatientMedication(models.Model):
    """
    Represents a specific medication prescribed or taken by a patient.

    Links a patient to a medication, its indication, dosage, intake times, and unit.
    """

    patient = models.ForeignKey("Patient", on_delete=models.CASCADE)
    medication_indication = models.ForeignKey("MedicationIndication", on_delete=models.CASCADE, related_name="indication_patient_medications", null=True)

    medication = models.ForeignKey("Medication", on_delete=models.CASCADE, blank=True, related_name="medication_patient_medications")

    intake_times = models.ManyToManyField(
        "MedicationIntakeTime",
        related_name="intake_time_patient_medications",
        blank=True,
    )

    unit = models.ForeignKey("Unit", on_delete=models.CASCADE, null=True, blank=True)
    dosage = models.JSONField(null=True, blank=True)
    active = models.BooleanField(default=True)

    objects = models.Manager()

    if TYPE_CHECKING:  # Added type hints block
        patient: models.ForeignKey["Patient"]
        medication_indication: models.ForeignKey["MedicationIndication|None"]
        medication: models.ForeignKey["Medication|None"]

        intake_times = cast(models.manager.RelatedManager["MedicationIntakeTime"], intake_times)
        unit: models.ForeignKey["Unit|None"]
        dosage = cast(models.JSONField, dosage)

    @property
    def links(self) -> "RequirementLinks":
        """
        Returns a RequirementLinks object for this PatientMedication instance.
        This is used during requirement evaluation.
        """
        from ....utils.links.requirement_link import RequirementLinks

        meds: List["Medication"] = []
        if self.medication:
            meds.append(self.medication)

        indications: List["MedicationIndication"] = []
        if self.medication_indication:
            indications.append(self.medication_indication)

        intake_times_list: List["MedicationIntakeTime"] = list(self.intake_times.all())

        return RequirementLinks(
            medications=meds,
            medication_indications=indications,
            medication_intake_times=intake_times_list,
            patient_medications=[self],  # Include self in patient_medications
        )

    class Meta:
        verbose_name = "Patient Medication"
        verbose_name_plural = "Patient Medications"

    @classmethod
    def create_by_patient_and_indication(cls, patient, medication_indication: "MedicationIndication"):
        """Creates a PatientMedication instance linking a patient and an indication."""

        patient_medication = cls.objects.create(patient=patient, medication_indication=medication_indication)
        patient_medication.save()

        return patient_medication

    def __str__(self):
        """Returns a string representation including medication, indication, dosage, and intake times."""
        intake_times = self.intake_times.all()
        out = f"{self.medication} (Indication {self.medication_indication}) - "
        out += f"{self.dosage} - {self.unit} - "

        for intake_time in intake_times:
            out += f"{intake_time} - "

        return out



# ReportReaderConfig Class
from typing import TYPE_CHECKING, cast

from django.db import models

if TYPE_CHECKING:
    from ....administration.center import Center
    from ....administration.person import FirstName, LastName
    from ....metadata import PdfType
    from .report_reader_flag import ReportReaderFlag


class ReportReaderConfig(models.Model):
    """
    Configuration settings for parsing PDF reports (ReportReader).

    Stores locale, name lists, date format, and flags used to identify key information lines
    and text sections to ignore.
    """

    locale = models.CharField(default="de_DE", max_length=10)
    first_names = models.ManyToManyField("FirstName", related_name="report_reader_configs")
    last_names = models.ManyToManyField("LastName", related_name="report_reader_configs")
    text_date_format = models.CharField(default="%d.%m.%Y", max_length=10)
    patient_info_line_flag = models.ForeignKey("ReportReaderFlag", related_name="report_reader_configs_patient_info_line", on_delete=models.CASCADE)
    endoscope_info_line_flag = models.ForeignKey("ReportReaderFlag", related_name="report_reader_configs_endoscope_info_line", on_delete=models.CASCADE)
    examiner_info_line_flag = models.ForeignKey("ReportReaderFlag", related_name="report_reader_configs_examiner_info_line", on_delete=models.CASCADE)
    cut_off_below = models.ManyToManyField("ReportReaderFlag", related_name="report_reader_configs_cut_off_below")
    cut_off_above = models.ManyToManyField("ReportReaderFlag", related_name="report_reader_configs_cut_off_above")

    if TYPE_CHECKING:
        patient_info_line_flag = models.ForeignKey["ReportReaderFlag"]
        endoscope_info_line_flag = models.ForeignKey["ReportReaderFlag"]
        examiner_info_line_flag = models.ForeignKey["ReportReaderFlag"]

        first_names = cast(models.manager.RelatedManager["FirstName"], first_names)
        last_names = cast(models.manager.RelatedManager["LastName"], last_names)
        cut_off_below = cast(models.manager.RelatedManager["ReportReaderFlag"], cut_off_below)
        cut_off_above = cast(models.manager.RelatedManager["ReportReaderFlag"], cut_off_above)

    def __str__(self):
        """Returns a string representation including the locale and primary key."""
        _str = f"ReportReaderConfig: {self.locale} (id: {self.pk}\n"
        return _str

    def update_names_by_center(self, center: "Center", save=True):
        """Updates the first and last name lists based on the names associated with a Center."""
        self.first_names.set(center.first_names.all())
        self.last_names.set(center.last_names.all())
        if save:
            self.save()

    def update_flags_by_pdf_type(self, pdf_type: "PdfType", save=True):
        """Updates the line identification flags based on a specific PdfType."""
        self.patient_info_line_flag = pdf_type.patient_info_line
        self.endoscope_info_line_flag = pdf_type.endoscope_info_line
        self.examiner_info_line_flag = pdf_type.examiner_info_line
        self.cut_off_below.set(pdf_type.cut_off_below_lines.all())
        self.cut_off_above.set(pdf_type.cut_off_above_lines.all())
        if save:
            self.save()



from typing import TYPE_CHECKING, List, Optional, Callable, Dict, Any
import re # Added import
from endoreg_db.models.medical.patient.patient_lab_value import PatientLabValue
from endoreg_db.models.requirement.requirement import Requirement
from datetime import datetime, timedelta

if TYPE_CHECKING:
    from endoreg_db.utils.links.requirement_link import RequirementLinks # Added import
    # from endoreg_db.models.requirement.requirement_operator import RequirementOperator # No longer directly used here

def get_latest_patient_lab_value(patient_lab_values: List[PatientLabValue], lab_value_name: str) -> Optional[PatientLabValue]:
    """
    Retrieves the most recent PatientLabValue for a specific lab_value_name.
    """
    relevant_values = [
        plv for plv in patient_lab_values if plv.lab_value and plv.lab_value.name == lab_value_name
    ]
    if not relevant_values:
        return None
    return max(relevant_values, key=lambda plv: plv.datetime)

def get_patient_lab_values_in_timeframe(
    patient_lab_values: List[PatientLabValue],
    lab_value_name: str,
    days_min: Optional[int] = None,
    days_max: Optional[int] = None,
) -> List[PatientLabValue]:
    """
    Retrieves PatientLabValues for a specific lab_value_name within a given timeframe.
    Timeframe is relative to the current date.
    days_min: e.g., -7 for 7 days ago (start of timeframe)
    days_max: e.g., 0 for today (end of timeframe)
    """
    relevant_values = [
        plv for plv in patient_lab_values if plv.lab_value and plv.lab_value.name == lab_value_name
    ]
    if not relevant_values:
        return []

    today = datetime.now().date()
    filtered_values = []

    for plv in relevant_values:
        plv_date = plv.datetime.date()
        in_timeframe = True
        if days_min is not None:
            start_date = today + timedelta(days=days_min)
            if plv_date < start_date:
                in_timeframe = False
        if days_max is not None:
            end_date = today + timedelta(days=days_max)
            if plv_date > end_date:
                in_timeframe = False
        if in_timeframe:
            filtered_values.append(plv)
    return filtered_values


def lab_latest_numeric_increased(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Returns True if the latest numeric lab value for all required lab values is above the normal maximum range; otherwise returns False.
    
    Returns False if any required lab value is missing, the latest value is unavailable, or the value does not exceed the normal maximum.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists():
        return False
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if not (latest_plv and latest_plv.value is not None and latest_plv.lab_value):
            return False
        patient_context = input_links.get_first_patient()
        normal_range = latest_plv.lab_value.get_normal_range(
            age=patient_context.age() if patient_context else None,
            gender=patient_context.gender if patient_context else None
        )
        if normal_range.get("max") is None or latest_plv.value <= normal_range["max"]:
            return False
    return True

def lab_latest_numeric_decreased(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Returns True if the latest numeric lab value for all required lab values is below the normal minimum range.
    
    Returns False if any latest value is missing, lacks a normal range, or is not below the minimum.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists(): # Changed
        return False
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if not (latest_plv and latest_plv.value is not None and latest_plv.lab_value): # Added check for latest_plv.lab_value
            return False
        patient_context = input_links.get_first_patient()
        normal_range = latest_plv.lab_value.get_normal_range(
            age=patient_context.age() if patient_context else None, 
            gender=patient_context.gender if patient_context else None
        )
        if normal_range.get("min") is None or latest_plv.value >= normal_range["min"]:
            return False
    return True

def lab_latest_numeric_normal(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Returns True if the latest numeric lab value for all required lab values is within the normal range.
    
    Returns False if any latest value is missing, lacks a normal range, or falls outside the normal range.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists(): # Changed
        return False
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if not (latest_plv and latest_plv.value is not None and latest_plv.lab_value): # Added check for latest_plv.lab_value
            return False
        patient_context = input_links.get_first_patient()
        normal_range = latest_plv.lab_value.get_normal_range(
            age=patient_context.age() if patient_context else None, 
            gender=patient_context.gender if patient_context else None
        )
        min_val = normal_range.get("min")
        max_val = normal_range.get("max")
        if min_val is None or max_val is None:
            return False
        if not (min_val <= latest_plv.value <= max_val):
            return False
    return True

def lab_latest_numeric_lower_than_value(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Returns True if the latest numeric lab value for all required lab values is lower than the specified threshold.
    
    Returns False if any latest value is missing or not below the requirement's numeric value.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists() or requirement.numeric_value is None: # Changed
        return False
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if not (latest_plv and latest_plv.value is not None):
            return False
        if not (latest_plv.value < requirement.numeric_value):
            return False
    return True

def lab_latest_numeric_greater_than_value(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Returns True if the latest numeric lab value for all required lab values is greater than the specified threshold.
    
    Returns False if any latest value is missing or not greater than the requirement's numeric value.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists() or requirement.numeric_value is None: # Changed
        return False
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if not (latest_plv and latest_plv.value is not None):
            return False
        if not (latest_plv.value > requirement.numeric_value):
            return False
    return True

# --- Operators with timeframe ---

def lab_latest_numeric_increased_factor_in_timeframe(
    input_links: "RequirementLinks", # Changed
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if the numeric lab value has increased by a certain factor within a timeframe.
    The factor is defined in requirement.numeric_value.
    The timeframe (in days relative to now) is defined by
    requirement.numeric_value_min (start, e.g., -7 for 7 days ago) and
    requirement.numeric_value_max (end, e.g., 0 for today).
    This operator checks if *any* value in the timeframe is increased by the factor compared to the value *at the start* of the timeframe.
    More sophisticated checks (e.g., sustained increase, comparison to earliest value in overall history) might require different logic.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.numeric_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    factor = requirement.numeric_value
    days_min = int(requirement.numeric_value_min) # Start of timeframe
    days_max = int(requirement.numeric_value_max) # End of timeframe

    for lab_value_model in requirement.lab_values.all():
        # Get all values for this lab type, not just within the timeframe initially
        all_lab_values_for_type = sorted(
            [plv for plv in patient_lab_values if plv.lab_value and plv.lab_value.name == lab_value_model.name and plv.value is not None],
            key=lambda plv: plv.datetime
        )

        if not all_lab_values_for_type:
            continue

        # Determine the reference value: the one closest to, but before or at, the start of the timeframe
        start_of_timeframe_date = (datetime.now() + timedelta(days=days_min)).date()
        reference_plv = None
        for plv in reversed(all_lab_values_for_type):
            if plv.datetime.date() <= start_of_timeframe_date:
                reference_plv = plv
                break
        
        if reference_plv is None and all_lab_values_for_type: # if no value before timeframe, use earliest available
             reference_plv = all_lab_values_for_type[0]


        if reference_plv and reference_plv.value is not None:
            reference_value = reference_plv.value
            # Now get values strictly within the defined timeframe
            values_in_timeframe = get_patient_lab_values_in_timeframe(
                patient_lab_values, lab_value_model.name, days_min, days_max
            )
            for plv_in_frame in values_in_timeframe:
                if plv_in_frame.value is not None and plv_in_frame.value > (reference_value * factor):
                    return True # Found a value increased by the factor
    return False

def lab_latest_numeric_decreased_factor_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if the numeric lab value has decreased by a certain factor within a timeframe.
    Factor is in requirement.numeric_value. Timeframe in numeric_value_min/max.
    Compares values in timeframe to the value at the start of (or just before) the timeframe.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.numeric_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    factor = requirement.numeric_value
    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        all_lab_values_for_type = sorted(
            [plv for plv in patient_lab_values if plv.lab_value and plv.lab_value.name == lab_value_model.name and plv.value is not None],
            key=lambda plv: plv.datetime
        )
        if not all_lab_values_for_type:
            continue

        start_of_timeframe_date = (datetime.now() + timedelta(days=days_min)).date()
        reference_plv = None
        for plv in reversed(all_lab_values_for_type):
            if plv.datetime.date() <= start_of_timeframe_date:
                reference_plv = plv
                break
        if reference_plv is None and all_lab_values_for_type:
            reference_plv = all_lab_values_for_type[0]

        if reference_plv and reference_plv.value is not None:
            reference_value = reference_plv.value
            if reference_value == 0: # Avoid division by zero or issues with zero reference
                continue 
            values_in_timeframe = get_patient_lab_values_in_timeframe(
                patient_lab_values, lab_value_model.name, days_min, days_max
            )
            for plv_in_frame in values_in_timeframe:
                if plv_in_frame.value is not None and plv_in_frame.value < (reference_value / factor): # Decreased by factor
                    return True
    return False

def lab_latest_numeric_normal_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if any numeric lab value within a timeframe is within its normal range.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)
    patient_context = input_links.get_first_patient()

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if plv.value is not None and plv.lab_value:
                normal_range = plv.lab_value.get_normal_range(
                    age=patient_context.age() if patient_context else None,
                    gender=patient_context.gender if patient_context else None
                )
                min_val = normal_range.get("min")
                max_val = normal_range.get("max")
                if min_val is not None and max_val is not None:
                    if min_val <= plv.value <= max_val:
                        return True
    return False

def lab_latest_numeric_lower_than_value_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if any numeric lab value within a timeframe is lower than requirement.numeric_value.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.numeric_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    threshold = requirement.numeric_value
    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if plv.value is not None and plv.value < threshold:
                return True
    return False

def lab_latest_numeric_greater_than_value_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if any numeric lab value within a timeframe is greater than requirement.numeric_value.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.numeric_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    threshold = requirement.numeric_value
    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if plv.value is not None and plv.value > threshold:
                return True
    return False

# --- Categorical Operators ---
# Assuming PatientLabValue has a 'value_string' attribute for categorical results.

def lab_latest_categorical_match(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if the latest categorical lab value matches requirement.string_value.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists() or requirement.string_value is None:
        return False

    match_string = requirement.string_value
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        # Ensure latest_plv has value_str attribute
        if latest_plv and hasattr(latest_plv, 'value_str') and latest_plv.value_str is not None: # Changed value_string to value_str
            if latest_plv.value_str == match_string:
                return True
    return False

def lab_latest_categorical_match_substring(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if requirement.string_value is a substring of the latest categorical lab value.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists() or requirement.string_value is None:
        return False

    substring = requirement.string_value
    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if latest_plv and hasattr(latest_plv, 'value_str') and latest_plv.value_str is not None: # Changed value_string to value_str
            if substring in latest_plv.value_str:
                return True
    return False

def lab_latest_categorical_match_regex(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if the latest categorical lab value matches regex in requirement.string_value.
    """
    patient_lab_values = input_links.patient_lab_values
    if not requirement.lab_values.exists() or requirement.string_value is None:
        return False

    regex_pattern = requirement.string_value
    try:
        compiled_regex = re.compile(regex_pattern)
    except re.error:
        return False # Invalid regex

    for lab_value_model in requirement.lab_values.all():
        latest_plv = get_latest_patient_lab_value(patient_lab_values, lab_value_model.name)
        if latest_plv and hasattr(latest_plv, 'value_str') and latest_plv.value_str is not None: # Changed value_string to value_str
            if compiled_regex.search(latest_plv.value_str):
                return True
    return False

# --- Categorical Operators with Timeframe ---

def lab_latest_categorical_match_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if any categorical lab value in timeframe matches requirement.string_value.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.string_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    match_string = requirement.string_value
    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if hasattr(plv, 'value_str') and plv.value_str is not None: # Changed value_string to value_str
                if plv.value_str == match_string:
                    return True
    return False

def lab_latest_categorical_match_substring_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if requirement.string_value is substring of any categorical lab value in timeframe.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.string_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    substring = requirement.string_value
    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if hasattr(plv, 'value_str') and plv.value_str is not None: # Changed value_string to value_str
                if substring in plv.value_str:
                    return True
    return False

def lab_latest_categorical_match_regex_in_timeframe(
    input_links: "RequirementLinks",
    requirement: Requirement,
    operator_kwargs: Dict[str, Any]
) -> bool:
    """
    Checks if any categorical lab value in timeframe matches regex in requirement.string_value.
    Timeframe in requirement.numeric_value_min/max.
    """
    patient_lab_values = input_links.patient_lab_values
    if (not requirement.lab_values.exists() or
        requirement.string_value is None or
        requirement.numeric_value_min is None or
        requirement.numeric_value_max is None):
        return False

    regex_pattern = requirement.string_value
    try:
        compiled_regex = re.compile(regex_pattern)
    except re.error:
        return False

    days_min = int(requirement.numeric_value_min)
    days_max = int(requirement.numeric_value_max)

    for lab_value_model in requirement.lab_values.all():
        values_in_timeframe = get_patient_lab_values_in_timeframe(
            patient_lab_values, lab_value_model.name, days_min, days_max
        )
        for plv in values_in_timeframe:
            if hasattr(plv, 'value_str') and plv.value_str is not None: # Changed value_string to value_str
                if compiled_regex.search(plv.value_str):
                    return True
    return False


# Mapping operator names to functions
LAB_VALUE_OPERATOR_FUNCTIONS: Dict[str, Callable] = {
    "lab_latest_numeric_increased": lab_latest_numeric_increased,
    "lab_latest_numeric_decreased": lab_latest_numeric_decreased,
    "lab_latest_numeric_normal": lab_latest_numeric_normal,
    "lab_latest_numeric_lower_than_value": lab_latest_numeric_lower_than_value,
    "lab_latest_numeric_greater_than_value": lab_latest_numeric_greater_than_value,

    # Aliases for backward compatibility, pointing to timeframe versions
    "lab_latest_numeric_increased_factor": lab_latest_numeric_increased_factor_in_timeframe,
    "lab_latest_numeric_decreased_factor": lab_latest_numeric_decreased_factor_in_timeframe,

    "lab_latest_numeric_increased_factor_in_timeframe": lab_latest_numeric_increased_factor_in_timeframe,
    "lab_latest_numeric_decreased_factor_in_timeframe": lab_latest_numeric_decreased_factor_in_timeframe,
    "lab_latest_numeric_normal_in_timeframe": lab_latest_numeric_normal_in_timeframe,
    "lab_latest_numeric_lower_than_value_in_timeframe": lab_latest_numeric_lower_than_value_in_timeframe,
    "lab_latest_numeric_greater_than_value_in_timeframe": lab_latest_numeric_greater_than_value_in_timeframe,

    "lab_latest_categorical_match": lab_latest_categorical_match,
    "lab_latest_categorical_match_substring": lab_latest_categorical_match_substring,
    "lab_latest_categorical_match_regex": lab_latest_categorical_match_regex,
    "lab_latest_categorical_match_in_timeframe": lab_latest_categorical_match_in_timeframe,
    "lab_latest_categorical_match_substring_in_timeframe": lab_latest_categorical_match_substring_in_timeframe,
    "lab_latest_categorical_match_regex_in_timeframe": lab_latest_categorical_match_regex_in_timeframe,
}



# Class to represent findings of examinations
from typing import TYPE_CHECKING, Optional, cast

from django.db import models


class FindingManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Finding(models.Model):
    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True, null=True)
    examinations = models.ManyToManyField("Examination", blank=True, related_name="findings")
    finding_types = models.ManyToManyField("FindingType", blank=True, related_name="findings")

    finding_interventions = models.ManyToManyField("FindingIntervention", blank=True, related_name="findings")

    objects = FindingManager()

    if TYPE_CHECKING:
        from endoreg_db.models import (
            Examination,
            FindingClassification,
            FindingClassificationType,
            FindingIntervention,
            FindingType,
            PatientFindingClassification,
        )

        finding_types = cast(models.manager.RelatedManager["FindingType"], finding_types)
        examinations = cast(models.manager.RelatedManager["Examination"], examinations)
        finding_interventions = cast(models.manager.RelatedManager["FindingIntervention"], finding_interventions)

        @property
        def finding_classifications(self) -> "models.manager.RelatedManager[FindingClassification]": ...

        # finding_classifications: models.QuerySet["FindingClassification"]
        # examinations: models.QuerySet[Examination]
        # finding_types: models.QuerySet[FindingType]
        # finding_interventions: models.QuerySet[FindingIntervention]

    def natural_key(self):
        """
        Return a tuple containing the unique natural key for this Finding instance.

        Returns:
            tuple: A single-element tuple with the Finding's name.
        """
        return (self.name,)

    def __str__(self):
        return self.name

    def get_finding_types(self):
        """
        Return all finding types associated with this finding.

        Returns:
            QuerySet: All related FindingType instances.
        """
        return self.finding_types.all()

    def get_classifications(self, classification_type: Optional[str] = None) -> models.QuerySet["FindingClassification"]:
        """
        Retrieve all classifications associated with this finding, optionally filtered by classification type.

        Parameters:
                classification_type (str, optional): If provided, only classifications with a matching type name are returned.

        Returns:
                List[FindingClassification]: List of related classification objects, filtered by type if specified.
        """
        if classification_type:
            return self.finding_classifications.filter(classification_types__name=classification_type)
        return self.finding_classifications.all()

    def get_location_classifications(self):
        """
        Retrieve all related FindingClassification objects with classification type 'location'.

        Returns:
            QuerySet: All FindingClassification instances linked to this finding where the classification type name is 'location' (case-insensitive).
        """
        return self.finding_classifications.filter(classification_types__name__iexact="location")

    def get_morphology_classifications(self):
        """
        Retrieve all related FindingClassification objects with classification type 'morphology'.

        Returns:
            QuerySet: A queryset of FindingClassification instances associated with this finding and classified as 'morphology'.
        """
        return self.finding_classifications.filter(classification_types__name__iexact="morphology")



"""Model for medication intake time"""

from typing import TYPE_CHECKING

from django.db import models


class MedicationIntakeTimeManager(models.Manager):
    """Manager for the medication intake time model."""

    def get_by_natural_key(self, name):
        """Retrieve a medication intake time by its natural key."""
        return self.get(name=name)


class MedicationIntakeTime(models.Model):
    """Model representing a medication intake time."""

    name = models.CharField(max_length=255, unique=True)  # Made unique=True
    repeats = models.CharField(max_length=20, default="daily")
    time = models.TimeField()

    objects = MedicationIntakeTimeManager()

    if TYPE_CHECKING:
        from endoreg_db.models import MedicationSchedule

        @property
        def medication_schedules(self) -> "models.manager.RelatedManager[MedicationSchedule]": ...

    @classmethod
    def dm(cls):
        """Return the daily morning intake time object."""
        return cls.objects.get(name="daily-morning")

    @classmethod
    def dno(cls):
        """Return the daily noon intake time object."""
        return cls.objects.get(name="daily-noon")

    @classmethod
    def de(cls):
        """Return the daily evening intake time object."""
        return cls.objects.get(name="daily-evening")

    @classmethod
    def dn(cls):
        """Return the daily night intake time object."""
        return cls.objects.get(name="daily-night")

    def natural_key(self):
        """Return the natural key for the medication intake time."""
        return (self.name,)

    def __str__(self):
        return self.name + " at " + str(self.time) + " (" + self.repeats + ")"



import string

def is_ascii(s):
    return all(c in string.printable for c in s)

def filter_non_ascii_lines(input_file, output_file):
    with open(input_file, 'r', encoding='utf-8') as infile:
        lines = infile.readlines()
    
    with open(output_file, 'w', encoding='utf-8') as outfile:
        for line in lines:
            if is_ascii(line):
                outfile.write(line)

if __name__ == "__main__":
    input_file = 'first_and_last_names_male.txt'
    output_file = 'first_and_last_names_male_ascii.txt'
    filter_non_ascii_lines(input_file, output_file)
    print(f"Filtered lines have been written to {output_file}")
